%%%  Matlab script file
%%%  Date: Jan, 5
%%%
%%% Investigate upsamping a 1D signal, introduce the convolution thm.

%%% Dependencies, Toolboxes:
%%%      iseToolbox/
%%%      320/matlab/utils/

%%% Author: ADJ
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%% Check Path and Constants  %%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
clear;
close all;

%%  Check path.  If these aren't on your path, then you will need
%%  to add these toolboxes to your path.  See ~jepson/pub/matlab/startup.m
which showIm          %% should be ...\iseToolbox\pyrTools\showIm.m

addpath('../utils');

TRUE = (1==1);
FALSE = ~TRUE;

TEST_SCRIPT = FALSE;  %% Set to true for testing path ok.

%%%%%%%%%%%  Initialize random number generator %%%%%%%%%%%%%%%%%%%%%%%
% Random number generator seed:
seed = round(sum(1000*clock));
seed = 2097255;  %% Hardwired seed ... comment out for new random samples
rand('state', seed);
randn('state', seed);
seed0 = seed;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%  Generate a 1D signal %%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

nSamp = 64;
nRange=0:(nSamp-1);   %% domain of signal S(n),  0 <= n < 64

%%%% Set prefilter
sigPre = 1.0;  %% sigma for Gaussian prefilter
filtK = round(2.5*sigPre);  %% prefilter radius
filtRange = -filtK:filtK;   %% prefilter support
preFilt = gauss1D(filtRange, 0.0, sigPre);  %% Gaussian prefilter
preFilt = preFilt/sum(preFilt);  %% normalize prefilt to sum to 1

%%%% Generate signal using prefilter
sigNoise = randn(1,nSamp);    %% white noise initial signal
%% Output signal is convolution of Gaussian prefilter with noise sample
signal = upConv(sigNoise, preFilt, 'circular');  


%%% Display signal, noise sample, prefilter
figure(1); clf;
hp(1)=plot(nRange, sigNoise, '-*r', 'LineWidth', 1.5, 'MarkerSize', 4);
hold on;
plot(nRange, zeros(size(nRange)), '-k', 'LineWidth', 1.5);
plot(nRange, signal, '-*b', 'LineWidth', 1.5, 'MarkerSize', 4);
plot(filtRange+round((nRange(end)+nRange(1))/2), preFilt, '-og',...
      'LineWidth', 1.5, 'MarkerSize', 4);
axis([0 nRange(end) -2.2 2.2]);
set(gca, 'FontSize', 14);
title('Signal (b), noise source (r), preFilter (g)');
xlabel('n');
ylabel('S');
%%print -depsc  signal

%%%%  In Fig. 1 the signal equals the noise signal smoothed 
%%%%  by the Gaussian prefilter
if TEST_SCRIPT
  pause(1);
else
  fprintf(1, '\nHit any key to continue...');
  pause;
  fprintf(1, '\n');
end
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%  Show the FT of the signal and source  %%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

FTsig = fft(signal);
FTsource = fft(sigNoise);
FTpreFilt = fft(embed(preFilt, size(signal)));

figure(2); clf; hold on;
plot(nRange, abs(FTsource), '-r*', 'LineWidth', 1.5, 'MarkerSize', 4);
plot(nRange, abs(FTsig), '-b*', 'LineWidth', 1.5, 'MarkerSize', 4);
plot(nRange, abs(FTpreFilt), '-go', 'LineWidth', 1.5, 'MarkerSize', 4);

axis([0 nRange(end) 0 18]);
set(gca, 'FontSize', 14);
title('Amplitude Spectrum (sig (b), src (r), flt (g))');
xlabel('k');
ylabel('|FT|');
%%print -depsc  ampSpec1Sig

%%%%  In Fig. 2 the amplitude spectrum of the signal equals the 
%%%% amp. spec. for the noise signal times the amp. spec. of the filter.
%%%% That is, these two signals should be essentially the same:
imStats(abs(FTsig), abs(FTsource) .* abs(FTpreFilt));

if TEST_SCRIPT
  pause(1);
else
  fprintf(1, '\nHit any key to continue...');
  pause;
  fprintf(1, '\n');
end
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%



%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%  Show the FT of the signal and source  %%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%  centered on frequency 0               %%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

FTsig = fftshift(fft(signal));
FTsource = fftshift(fft(sigNoise));
FTpreFilt = fftshift(fft(embed(preFilt, size(signal))));

kRange = floor(length(signal)/2);
kRange = -kRange:1:kRange;
kRange = kRange(1:length(signal));

figure(2); clf; hold on;
plot(kRange, abs(FTsource), '-r*', 'LineWidth', 1.5, 'MarkerSize', 4);
plot(kRange, abs(FTsig), '-b*', 'LineWidth', 1.5, 'MarkerSize', 4);
plot(kRange, abs(FTpreFilt), '-go', 'LineWidth', 1.5, 'MarkerSize', 4);

axis([kRange(1) kRange(end) 0 18]);
set(gca, 'FontSize', 14);
title('Amplitude Spectrum (sig (b), src (r), flt (g))');
xlabel('k');
ylabel('|FT|');
%%print -depsc  ampSpecSig

if TEST_SCRIPT
  pause(1);
else
  fprintf(1, '\nHit any key to continue...');
  pause;
  fprintf(1, '\n');
end
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%  Upsample the signal by a factor of 3  %%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

upRate = 3;
%% Set signal to be all zeros.
upSig = zeros(1, length(signal)*upRate);
%% Set every upRate pixel equal to a signal value.
upSig(1+upRate*nRange) = signal;
nUpRange = 0:(upRate*length(nRange)-1);

figure(1); clf;
plot(nUpRange, upSig, '-*r', 'LineWidth', 1.5, 'MarkerSize', 4);
ax = axis;
mx = max(abs(ax(3:4)));
axis([0 nUpRange(end) -mx mx ]);
set(gca, 'FontSize', 14);
title('Raw Upsampled Signal');
xlabel('n');
ylabel('S');
%%print -depsc  upSig

if TEST_SCRIPT
  pause(1);
else
  fprintf(1, '\nHit any key to continue...');
  pause;
  fprintf(1, '\n');
end
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%% FT of up sampled signal
kRange = floor(length(nUpRange)/2);
kRange = -kRange:1:kRange;
kRange = kRange(1:length(nUpRange));

figure(2); clf;
plot(kRange, abs(fftshift(fft(upSig))), '-r*', ...
     'LineWidth', 1.5, 'MarkerSize', 4);
ax = axis;
mx = max(abs(ax(3:4)));
axis([ax(1) ax(2) 0 mx ]);
set(gca, 'FontSize', 14);
title('Amplitude Spectrum of Raw Upsampled Signal');
xlabel('k');
ylabel('|FT|');
%%print -depsc  FTupSigRaw

if TEST_SCRIPT
  pause(1);
else
  fprintf(1, '\nHit any key to continue...');
  pause;
  fprintf(1, '\n');
end
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%  Filter Upsampled Signal %%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%% Triangular Kernel
tFilt = [1:upRate (upRate-1):-1:1]/upRate;
kFilt = floor(length(tFilt)/2);

figure(1);clf;
plot((-kFilt-1):(kFilt+1), [0 tFilt 0],'-*g', ...
     'LineWidth', 1.5, 'MarkerSize', 4);
ax = axis;
mx = max(abs(ax(3:4)));
axis([ax(1) ax(2) 0 mx ]);
set(gca, 'FontSize', 14);
title('Triangular Filter');
xlabel('k');
ylabel('Filt');
%%print -depsc  triangFilt

if TEST_SCRIPT
  pause(1);
else
  fprintf(1, '\nHit any key to continue...');
  pause;
  fprintf(1, '\n');
end
%%%%


%%%% Convolve triangular kernel with upsampled signal
newSig = upConv(upSig, tFilt, 'circular');

figure(1);clf;
plot(upSig,'-*r', 'LineWidth', 1.5, 'MarkerSize', 4);
hold on;
plot(newSig,'-*b');
ax = axis;
mx = max(abs(ax(3:4)));
axis([0 nUpRange(end) -mx mx ]);
set(gca, 'FontSize', 14);
title('Triang-Filtered Upsampled Signal (b)');
xlabel('n');
ylabel('S');
%%print -depsc  triangFiltUpSamp

if TEST_SCRIPT
  pause(1);
else
  fprintf(1, '\nHit any key to continue...');
  pause;
  fprintf(1, '\n');
end
%%%%%

%%%% Show FFT of filtered upsampled signal
kRange = floor(length(nUpRange)/2);
kRange = -kRange:1:kRange;
kRange = kRange(1:length(nUpRange));
figure(1); clf;
plot(kRange, upRate*abs(fftshift(fft(upSig))), '-r*',...
     'LineWidth', 1.5, 'MarkerSize', 4);
tmp = embed(tFilt, size(newSig));
hold on;
plot(kRange, abs(fftshift(fft(tmp))), '-g*', ...
     'LineWidth', 1.5, 'MarkerSize', 4);
plot(kRange, abs(fftshift(fft(newSig))), '-b*',...
     'LineWidth', 1.5, 'MarkerSize', 4);
     
ax = axis;
mx = max(abs(ax(3:4)));
axis([kRange(1) kRange(end) 0 mx ]);
set(gca, 'FontSize', 14);
title('Amp. Spec. (Up. Sig.(r), Trg. Flt. (g), Flt. Sig.(b)');
xlabel('k');
ylabel('|FT|');
%%print -depsc  FTtriangFiltUpSamp

if TEST_SCRIPT
  pause(1);
else
  fprintf(1, '\nHit any key to continue...');
  pause;
  fprintf(1, '\n');
end
%%%%%

%%%%%%% Use a Catmull-Rom Filter %%%%%%%%%%
crFilt = catmullRom(upRate);

kRad = floor(length(crFilt)/2);
figure(1);clf;
plot((-kRad-1):(kRad+1), [0 crFilt 0],'-*g', ...
     'LineWidth', 1.5, 'MarkerSize', 4);
ax = axis;
hold on;
plot([ax(1) ax(2)], zeros(1,2), '-k');
axis(ax);
set(gca, 'FontSize', 14);
title('Catmull-Rom Filter');
xlabel('k');
ylabel('Filt');
%%print -depsc  CRFilt

if TEST_SCRIPT
  pause(1);
else
  fprintf(1, '\nHit any key to continue...');
  pause;
  fprintf(1, '\n');
end
%%%%


%%%% Convolve Catmull-Rom kernel with upsampled signal
newSig = upConv(upSig, crFilt, 'circular');

figure(1);clf;
plot(upSig,'-*r', 'LineWidth', 1.5, 'MarkerSize', 4);
hold on;
plot(newSig,'-*b');
ax = axis;
mx = max(abs(ax(3:4)));
axis([0 nUpRange(end) -mx mx ]);
set(gca, 'FontSize', 14);
title('Catmull-Rom Filtered Upsampled Signal (b)');
xlabel('n');
ylabel('S');
%%print -depsc  crFiltUpSamp

if TEST_SCRIPT
  pause(1);
else
  fprintf(1, '\nHit any key to continue...');
  pause;
  fprintf(1, '\n');
end
%%%%%

%%%% Show FFT of filtered upsampled signal
kRange = floor(length(nUpRange)/2);
kRange = -kRange:1:kRange;
kRange = kRange(1:length(nUpRange));
figure(1); clf;
plot(kRange, upRate*abs(fftshift(fft(upSig))), '-r*',...
     'LineWidth', 1.5, 'MarkerSize', 4);
tmp = embed(crFilt, size(newSig));
hold on;
plot(kRange, abs(fftshift(fft(tmp))), '-g*', ...
     'LineWidth', 1.5, 'MarkerSize', 4);
plot(kRange, abs(fftshift(fft(newSig))), '-b*',...
     'LineWidth', 1.5, 'MarkerSize', 4);
     
ax = axis;
mx = max(abs(ax(3:4)));
axis([kRange(1) kRange(end) 0 mx ]);
set(gca, 'FontSize', 14);
title('Amp. Spec. (Up. Sig.(r), CR Flt. (g), Flt. Sig.(b)');
xlabel('k');
ylabel('|FT|');
%%print -depsc  FTcrFiltUpSamp

if TEST_SCRIPT
  pause(1);
else
  fprintf(1, '\nHit any key to continue...');
  pause;
  fprintf(1, '\n');
end
%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%  Sinc Filter Upsampled Signal %%%%%%%%%%%%%%%%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% Alternatively, we can specify an "ideal" filter in frequency
%% space, which passes only the low-frequencies in the up sampled
%% signal upSig.

%% Define an ideal low-pass filter.
kRange = floor(length(nUpRange)/2);
kRange = -kRange:1:kRange;
kRange = kRange(1:length(nUpRange));
kLow = -floor(nSamp/2);
kHigh = kLow + (nSamp-1) + mod(nSamp,2);  %% Make sure kernel has an odd length
idx = ((kRange(:) >= kLow ) & (kRange(:) <= kHigh));
FTsinc = zeros(size(upSig));
FTsinc(idx) = upRate;

%% Look at the corresponding filter
sincFilt = real(ifft(fftshift(FTsinc)));
sincFilt = fftshift(sincFilt);

kRange = floor(length(nUpRange)/2);
kRange = -kRange:1:kRange;
kRange = kRange(1:length(nUpRange));

figure(1);clf;
plot(kRange, sincFilt,'-*g', ...
     'LineWidth', 1.5, 'MarkerSize', 4);
ax = axis;
mx = max(abs(ax(3:4)));
%axis([ax(1) ax(2) -mx mx ]);
set(gca, 'FontSize', 14);
title('Sinc Filter');
xlabel('k');
ylabel('Filt');
%%print -depsc  sincFilt

if TEST_SCRIPT
  pause(1);
else
  fprintf(1, '\nHit any key to continue...');
  pause;
  fprintf(1, '\n');
end
%%%%

%%%% Convolve sinc kernel with upsampled signal
newSig = upConv(sincFilt, upSig, 'circular');

figure(1);clf;
plot(upSig,'-*r', 'LineWidth', 1.5, 'MarkerSize', 4);
hold on;
plot(newSig,'-*b');
ax = axis;
mx = max(abs(ax(3:4)));
axis([0 nUpRange(end) -mx mx ]);
set(gca, 'FontSize', 14);
title('Sinc Filtered Upsampled Signal (b)');
xlabel('n');
ylabel('S');
%%print -depsc  sincFiltUpSamp

if TEST_SCRIPT
  pause(1);
else
  fprintf(1, '\nHit any key to continue...');
  pause;
  fprintf(1, '\n');
end
%%%%%

%%%% Show FFT of filtered upsampled signal
kRange = floor(length(nUpRange)/2);
kRange = -kRange:1:kRange;
kRange = kRange(1:length(nUpRange));
figure(1); clf;
plot(kRange, upRate*abs(fftshift(fft(upSig))), '-r*',...
     'LineWidth', 1.5, 'MarkerSize', 4);
tmp = embed(sincFilt, size(newSig));
hold on;
plot(kRange, abs(fftshift(fft(tmp))), '-g*', ...
     'LineWidth', 1.5, 'MarkerSize', 4);
plot(kRange, abs(fftshift(fft(newSig))), '-b*',...
     'LineWidth', 1.5, 'MarkerSize', 4);
     
ax = axis;
mx = max(abs(ax(3:4)));
axis([kRange(1) kRange(end) 0 mx ]);
set(gca, 'FontSize', 14);
title('Amp. Spec. (Upsampled Sig.(r), Sinc Filter(g), Filtered Sig.(b)');
xlabel('k');
ylabel('|FT|');
%%print -depsc  FTsincFiltUpSamp

fprintf(1, '\nThe End\n');
%%%%%
